#include <config.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#else
#include <glib.h>
#endif

#include <glib/gstdio.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "main.h"

#include <taglib/tag_c.h>

#include "bmpx/util.h"
#include "bmpx/file_utils.h"
#include "bmpx/metadata.h"
#include "bmpx/uri.h"


G_DEFINE_TYPE (BmpMetadataCache, bmp_metadata_cache, G_TYPE_OBJECT)

enum {
    BMP_METADATA_CACHE_SIGNAL_FILE_ADDED,

    BMP_METADATA_N_SIGNALS
};
static guint signals[BMP_METADATA_N_SIGNALS] = {0};

enum {
    LIST_ARTIST,
    LIST_ALBUM
};

struct _BmpMetadataCachePrivate {
        gboolean dispose_has_run;

	FILE *fh;

	gint ufid;

	GHashTable *ht_uri;
	GHashTable *ht_ufid;

	/* This is sort of a hack for the current implementation to build the tree view fast */
	GHashTable *ht_artist_albums;
	GHashTable *ht_album_songs;

	/* We hold a shallow-copy-based hash table for each metadatum */
	GHashTable *ht_metadatum[BMP_DATUM_NONE];
};

struct _BmpTagDefine {
    gchar *title;
    gchar *id;

    BmpMetadatumId tag_id;
};
typedef struct _BmpTagDefine BmpTagDefine;

/* We're using GST-style tag identifiers here */
static BmpTagDefine bmp_tag_defines[] = {
    {"Title",             "title",          BMP_DATUM_TITLE},
    {"Time",              "time",           BMP_DATUM_TIME},
    {"Track",             "track-number",   BMP_DATUM_TRACK},
    {"Rating",            "rating",         BMP_DATUM_RATING},
    {"Artist",            "artist",         BMP_DATUM_ARTIST},
    {"Album",             "album",          BMP_DATUM_ALBUM},
    {"Date",              "date",           BMP_DATUM_DATE},
    {"Genre",             "genre",          BMP_DATUM_GENRE},
    {"Comment",           "comment",        BMP_DATUM_COMMENT},
    {"Bitrate",           "bitrate",        BMP_DATUM_BITRATE},
    {"Location",          "location",       BMP_DATUM_LOCATION},
    {"MTIME",		  "mtime",	    BMP_DATUM_MTIME},
    {"Samplerate",        "samplerate",     BMP_DATUM_SAMPLERATE},
};

G_LOCK_DEFINE(METADATA);

gchar*
bmp_metadata_get_id (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].id);
}

const gchar*
bmp_metadata_get_id_static (BmpMetadatumId tag_id)
{
    return bmp_tag_defines[tag_id].id;
}


gchar*
bmp_metadata_get_title (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].title);
}

const gchar*
bmp_metadata_get_metadata_field (GHashTable *metadata, BmpMetadatumId tag_id)
{
	gchar *id_str, *rv;


	id_str =
	    bmp_metadata_get_id(tag_id);
	rv =
	    g_hash_table_lookup (metadata, id_str);

	g_free (id_str);

	return rv;
}

static void
metadata_get_from_uri (GHashTable *metadata, const gchar *uri)
{
    TagLib_File *file =
        NULL;

    TagLib_Tag *tag =
        NULL;

    TagLib_AudioProperties const *audio =
        NULL;

    gchar *artist,
          *album,
          *title,
          *track,
          *year,
          *genre,
          *comment,
          *ttime,
          *rating,
          *bitrate,
	  *samplerate,
          *title_rep = NULL,
          *filename,
          *mtime_string;

    GURI *g_uri;
    struct stat file_stat;

    g_uri =
	bmp_uri_new (uri);
    bmp_uri_unescape (g_uri);

    filename =
	g_uri->path;

    if (!filename)
      {
	bmp_uri_delete (g_uri);
	return;
      }

    /* store mtime */
    stat (filename, &file_stat);
    mtime_string =
	g_strdup_printf("%li", file_stat.st_mtime);

    g_hash_table_insert (metadata,
                              bmp_metadata_get_id
                              (BMP_DATUM_MTIME), mtime_string);

    /* store location (uri) */
    g_hash_table_insert (metadata,
                              bmp_metadata_get_id
                             (BMP_DATUM_LOCATION), g_strdup(uri));

    file =
        taglib_file_new (filename);

    if (file)
     {
         tag =
             taglib_file_tag (file);

         if (!tag)
          {
              g_message ("%s: no metadata", filename);
	      taglib_file_free (file);
              return;
          }

         audio =
             taglib_file_audioproperties (file);

         track =
             g_strdup_printf ("%d", taglib_tag_track (tag));
         year =
             g_strdup_printf ("%d", taglib_tag_year (tag));
         artist =
             taglib_tag_artist (tag);
         album =
             taglib_tag_album (tag);
         title =
             taglib_tag_title (tag);
         genre =
             taglib_tag_genre (tag);
         comment =
             taglib_tag_comment (tag);
         ttime =
             g_strdup_printf ("%d", taglib_audioproperties_length (audio));
         rating =
             g_strdup ("0");
         bitrate =
             g_strdup_printf ("%d",taglib_audioproperties_bitrate (audio));
	 samplerate =
             g_strdup_printf ("%d",taglib_audioproperties_samplerate (audio));

         g_hash_table_insert (metadata,
                              bmp_metadata_get_id
                              (BMP_DATUM_SAMPLERATE), samplerate);
         g_hash_table_insert (metadata,
                              bmp_metadata_get_id
                              (BMP_DATUM_BITRATE), bitrate);
         g_hash_table_insert (metadata,
                              bmp_metadata_get_id
                              (BMP_DATUM_RATING), rating);
         g_hash_table_insert (metadata,
                              bmp_metadata_get_id (BMP_DATUM_TIME),
                              ttime);
         g_hash_table_insert (metadata,
                              bmp_metadata_get_id (BMP_DATUM_TRACK),
                              track);
         g_hash_table_insert (metadata,
                              bmp_metadata_get_id (BMP_DATUM_ARTIST),
                              ((!artist) || (!strlen(artist))) ? g_strdup("(Unknown Artist)") : g_strdup (artist));

	if (strlen (title) == 0)
          {
              title_rep = g_path_get_basename (filename);
              g_hash_table_insert (metadata,
                                   bmp_metadata_get_id
                                   (BMP_DATUM_TITLE), g_strdup (title_rep));
              g_free (title_rep);
          }
         else
          {
              g_hash_table_insert (metadata,
                                   bmp_metadata_get_id
                                   (BMP_DATUM_TITLE), g_strdup (title));
          }

#if 0
         if (strlen (album) == 0)
          {
              gchar **path_frags;
              gint n = 0;

              path_frags = g_strsplit (filename, "/", -1);

              while (path_frags[n + 1])
               {
                   album_rep = path_frags[n++];
               }

              g_hash_table_insert (metadata,
                                   bmp_metadata_get_id
                                   (BMP_DATUM_ALBUM),
                                   g_strdup (album_rep));
              g_strfreev (path_frags);
          }
         else
          {
#endif
              g_hash_table_insert (metadata,
                                   bmp_metadata_get_id
                                   (BMP_DATUM_ALBUM),
				   ((!album) || (!strlen(album))) ? g_strdup("(Unknown Album)") : g_strdup (album));
#if 0
          }
#endif

         g_hash_table_insert (metadata,
                              bmp_metadata_get_id (BMP_DATUM_GENRE),
                              g_strdup (genre));
         g_hash_table_insert (metadata,
                              bmp_metadata_get_id (BMP_DATUM_DATE),
                              year);
         g_hash_table_insert (metadata,
                              bmp_metadata_get_id
                              (BMP_DATUM_COMMENT), g_strdup (comment));

         taglib_tag_free_strings ();
         taglib_file_free (file);
     }

    bmp_uri_delete (g_uri);

}

typedef struct {
    gint start;
    gint end;
    BmpListType *list;
} MetadataThreadTuple;


static gpointer
metadata_thread (MetadataThreadTuple *mtt)
{
   BmpFileItem *item;
   gint n;

   for (n = mtt->start; n < mtt->end; n++)
   {
	item = g_list_nth(mtt->list->items, n)->data;

	if (item->metadata == NULL)
	{
	    item->locked = TRUE;
	    item->metadata = bmp_metadata_cache_get_metadata_for_uri (bmp_metadata_cache, item->uri);
	    item->dirty = TRUE;
	    item->locked = FALSE;
	}
   }

    g_thread_exit(NULL);

    /* Shut GCC up */
    return NULL;
}

#if 0
void
bmp_metadata_get_for_files_threaded (BmpListType *list, gint start, gint end)
{
    GThread *mt;
    GError *error = NULL;
    MetadataThreadTuple *mtt;

    if (!list) return;
    if (!list->items) return;

    mtt = g_new0(MetadataThreadTuple, 1);

    mtt->start = start;
    mtt->end = end;
    mtt->list = list;

    mt = g_thread_create ((GThreadFunc) metadata_thread, mtt, TRUE, &error);
}
#endif

/* Media Cache functions */
static void
bmp_metadata_cache_init (BmpMetadataCache *self)
{
	gint n;

        self->private = g_new (BmpMetadataCachePrivate, 1);

        self->private->dispose_has_run = FALSE;

	self->private->ufid = -1;

	self->private->ht_uri = g_hash_table_new (g_str_hash, g_str_equal);
	self->private->ht_ufid = g_hash_table_new (g_direct_hash, g_direct_equal);

	for (n = 0; n < BMP_DATUM_NONE;
	    self->private->ht_metadatum[n++] = g_hash_table_new (g_str_hash, g_str_equal));

	self->private->ht_artist_albums = g_hash_table_new (g_str_hash, g_str_equal);
	self->private->ht_album_songs = g_hash_table_new (g_str_hash, g_str_equal);

	g_message("%s\t\t[INITIALIZED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static void
bmp_metadata_cache_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
bmp_metadata_cache_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpMetadataCacheClass *klass;
                GObjectClass *parent_class;
                klass = BMP_METADATA_CACHE_CLASS (g_type_class_peek (BMP_TYPE_METADATA_CACHE));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
bmp_metadata_cache_dispose (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        if (self->private->dispose_has_run) {
                return;
        }

	bmp_metadata_cache_write (self);

        self->private->dispose_has_run = TRUE;

	g_message("%s\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_finalize (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        g_free (self->private);
}

BmpMetadataCache*
bmp_metadata_cache_new (void)
{
    BmpMetadataCache *self;

    self = g_object_new (bmp_metadata_cache_get_type (), NULL);
    bmp_metadata_cache_read (self);

    return self;
}

static void
bmp_metadata_cache_class_init (BmpMetadataCacheClass *g_class)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

        gobject_class->set_property = bmp_metadata_cache_set_property;
        gobject_class->get_property = bmp_metadata_cache_get_property;
        gobject_class->dispose = bmp_metadata_cache_dispose;
        gobject_class->finalize = bmp_metadata_cache_finalize;
        gobject_class->constructor = bmp_metadata_cache_constructor;

        signals[BMP_METADATA_CACHE_SIGNAL_FILE_ADDED] =
	    g_signal_new ("file_added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, file_added),
                      NULL, NULL, g_cclosure_marshal_VOID__INT,
                      G_TYPE_NONE, 1, G_TYPE_INT);

}

/* Public API */
GHashTable*
bmp_metadata_cache_get_metadata_for_ufid (BmpMetadataCache *self, gint ufid)
{
    return g_hash_table_lookup (self->private->ht_ufid, GINT_TO_POINTER(ufid));
}

const gchar*
bmp_metadata_cache_get_metadata_field_from_ufid (BmpMetadataCache *self, gint ufid, BmpMetadatumId tag_id)
{
    GHashTable *ht_aux;

    ht_aux = g_hash_table_lookup (self->private->ht_ufid, GINT_TO_POINTER(ufid));

    return g_hash_table_lookup (ht_aux, bmp_metadata_get_id_static(tag_id));
}

GHashTable*
bmp_metadata_cache_get_metadata_for_uri (BmpMetadataCache *self, const gchar *uri)
{
    GHashTable *metadata = NULL;

    G_LOCK(METADATA);

    metadata =
	g_hash_table_lookup (self->private->ht_uri, uri);

    if (! metadata)
    {
	gchar *ufid;
	gint n;

	if (!g_ascii_strncasecmp("file",uri,4))
	  {
		metadata =
		    g_hash_table_new (g_str_hash, g_str_equal);
		metadata_get_from_uri (metadata, uri);
	  }
	else
	  {
	        metadata =
		    g_hash_table_new (g_str_hash, g_str_equal);
	        g_hash_table_insert (metadata, bmp_metadata_get_id(BMP_DATUM_LOCATION), g_strdup(uri));
#ifdef HAVE_XINE
		bmp_play_get_metadata_xine (bmp_play, metadata, uri);
#endif
#ifdef HAVE_GST
		bmp_play_get_metadata_gst (bmp_play, uri);
#endif
	  }

	self->private->ufid++;

	ufid = g_strdup_printf("%d", self->private->ufid);
	g_hash_table_insert (metadata, g_strdup("ufid"), ufid);

	g_hash_table_insert (self->private->ht_ufid, GINT_TO_POINTER(ufid), metadata);
	g_hash_table_insert (self->private->ht_uri,
		    g_strdup(g_hash_table_lookup(metadata, bmp_metadata_get_id_static(BMP_DATUM_LOCATION))), metadata);

	for (n = 0; n < BMP_DATUM_NONE; n++)
	  {
	    GList *ufid_list = NULL;
	    gchar *key;

	    /* We acquire the particular metadata value from the just added file */
	    key = g_hash_table_lookup (metadata, bmp_metadata_get_id_static(n));

	    if (key)
	      {
		/* Now let's see if a list for this particular datum (Artist, Album...) already exists */
		ufid_list = g_hash_table_lookup (self->private->ht_metadatum[n], key);

		if (! ufid_list) /*..nope*/
                  {
		     ufid_list = g_list_append (ufid_list, GINT_TO_POINTER(ufid));
		     g_hash_table_insert (self->private->ht_metadatum[n], key, ufid_list);
		  }
		else
                  {
		     ufid_list = g_list_append (ufid_list, GINT_TO_POINTER(ufid));
		     g_hash_table_replace (self->private->ht_metadatum[n], key, ufid_list);
		  }
	      }
	  }


	g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_FILE_ADDED], 0, self->private->ufid);

    } else if (G_URI_IS_LOCAL(uri)) {

	struct stat file_stat;
	long int st_mtime_cmp;
	GURI *g_uri;

	st_mtime_cmp =
	    strtol (g_hash_table_lookup (metadata, bmp_metadata_get_id_static (BMP_DATUM_MTIME)), NULL, 10);
	g_uri =
	    bmp_uri_new (uri);

	stat (g_uri->path, &file_stat);

	if (st_mtime_cmp != file_stat.st_mtime) {
	    /* FIXME: Action on file has changed */
	}

	bmp_uri_delete (g_uri);
    }

    G_UNLOCK(METADATA);

    return metadata;
}

GList*
bmp_metadata_cache_get_albums_for_artist (BmpMetadataCache *self,
					  const gchar *artist)
{
    BmpUniqueList *albums;

    albums = g_hash_table_lookup (self->private->ht_artist_albums, artist);

    if (! albums)
	return NULL;

    return bmp_unique_list_to_glist (albums);
}

GList*
bmp_metadata_cache_get_songs_for_album (BmpMetadataCache *self,
					const gchar *album)
{
    BmpUniqueList *songs;

    songs = g_hash_table_lookup (self->private->ht_album_songs, album);

    if (!songs)
	return NULL;

    return bmp_unique_list_to_glist (songs);
}

#define PACKET_LIBNAME   (uint8_t)0x01
#define PACKET_MEDIAFILE (uint8_t)0x02
#define PACKET_EMC       (uint8_t)0x03

#define PACKET_MF_UFID      (uint8_t)0x01
#define PACKET_MF_URI       (uint8_t)0x02
#define PACKET_MF_MVALUE    (uint8_t)0x03
#define PACKET_MF_EMF       (uint8_t)0x04

static void
ht_metadata_foreach (gpointer key, gpointer data, BmpMetadataCache *self)
{
    FILE *fh;
    BmpMetadatumId tag_id = -1;
    gint n;
    uint8_t value;
    uint16_t value16;
    uint8_t packet_id;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    if (! strcmp(key, "ufid")) return;
    if (! data) return;

    for (n = 0; n < BMP_DATUM_NONE; n++)
    {
	if (!strcmp(key, bmp_metadata_get_id_static(n)))
	{
	    tag_id = n;
	    break;
	}
    }

    if (tag_id == -1) return;

    fh = self->private->fh;

    packet_id = PACKET_MF_MVALUE;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);

    value = tag_id;
    fwrite (&value, sizeof(uint8_t), 1, fh);

    value16 = strlen(data);
    fwrite (&value16, sizeof(uint16_t), 1, fh);
    fwrite (data, sizeof(char), strlen(data), fh);
}

static void
ht_uri_foreach (gpointer key, gpointer data, BmpMetadataCache *self)
{
    GHashTable *metadata;
    FILE *fh;
    uint8_t  packet_id;
    uint16_t value;
    gchar *filename = NULL;
    GURI *g_uri;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    g_uri =
	bmp_uri_new (key);
    filename =
	g_uri->path;

    if (!g_ascii_strncasecmp("file",key,4))
      {
	bmp_uri_unescape (g_uri);
	/* File is not present anymore, discard */
	if (!g_file_test (g_uri->path, G_FILE_TEST_EXISTS))
	  {
	    bmp_uri_delete (g_uri);
	    return;
	  }
      }

    metadata = (GHashTable*)data;
    fh = self->private->fh;

    /* Write MF identification */
    packet_id = PACKET_MEDIAFILE;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);

    /* Write UFID */
    packet_id = PACKET_MF_UFID;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);
    value = strtol(g_hash_table_lookup(metadata,"ufid"), NULL, 10);
    fwrite (&value, sizeof(uint16_t), 1, fh);

    /* Write URI */
    packet_id = PACKET_MF_URI;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);

    value = (uint16_t)strlen(key);
    fwrite (&value, sizeof(uint16_t), 1, fh);
    fwrite (key, sizeof(char), strlen(key), fh);

    /* Now write all metadata */
    g_hash_table_foreach (metadata, (GHFunc) ht_metadata_foreach, self);

    /* Write EMF */
    packet_id = PACKET_MF_EMF;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);

    bmp_uri_delete (g_uri);
}

void
bmp_metadata_cache_write (BmpMetadataCache *self)
{
    FILE *fh;
    static const gchar *mcid = "MC";
    static const uint16_t mcidn = 666;
    uint8_t value;
    gchar *path;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    /* We make the operation atomic by writing to a temp file first and only if we
       fully succeed we move the temp to the actual file */

    path = g_strconcat (bmp_paths[BMP_PATH_METADATA_CACHE], ".tmp", NULL);

    self->private->fh = g_fopen(path, "w");
    fh = self->private->fh;

    /* Write file magic */
    fwrite (mcid, sizeof(char), strlen(mcid), fh);
    fwrite (&mcidn, sizeof(uint16_t), 1, fh);

    /* Process cache */
    g_hash_table_foreach (self->private->ht_uri, (GHFunc) ht_uri_foreach, self);

    /* Write EMC */
    value = PACKET_EMC;
    fwrite (&value, sizeof(uint8_t), 1, fh);
    fclose (self->private->fh);

    g_remove (bmp_paths[BMP_PATH_METADATA_CACHE]);
    g_rename (path, bmp_paths[BMP_PATH_METADATA_CACHE]);

#if 0
    g_message("%s: Write complete.", G_STRFUNC);
#endif
}

void
bmp_metadata_cache_read (BmpMetadataCache *self)
{
    FILE *fh;
    static gchar *mcid;
    static uint16_t mcidn;
    uint8_t packet_id;
    uint8_t value;
    uint16_t value16;
    char *valuestr;
    gboolean run = TRUE;
    gint n;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    if (!g_file_test(bmp_paths[BMP_PATH_METADATA_CACHE], G_FILE_TEST_EXISTS)) return;

    self->private->fh = g_fopen(bmp_paths[BMP_PATH_METADATA_CACHE], "r");
    fh = self->private->fh;

    /* Write file magic */
    mcid = g_malloc0(3);
    fread (mcid, sizeof(char), 2, fh);
    fread (&mcidn, sizeof(uint16_t), 1, fh);

    if (strcmp(mcid,"MC")) {
	g_message("%s: Wrong file magic of metadata cache file %s",
		  G_STRFUNC, bmp_paths[BMP_PATH_METADATA_CACHE]);
	fclose(self->private->fh);
	return;
    }

    if (mcidn != 666) {
	g_message("%s: Wrong file magic for metadata cache file %s",
		  G_STRFUNC, bmp_paths[BMP_PATH_METADATA_CACHE]);
	fclose(self->private->fh);
	return;
    }

    /* Read the metadata cache */
    while (run)
    {
	    fread (&packet_id, sizeof(uint8_t), 1, fh);

	    switch (packet_id)
	    {
		/* Currently unsupported */
		case PACKET_LIBNAME:
		    break;

		case PACKET_EMC:
		    {
			run = FALSE;
			break;
		    }

		case PACKET_MEDIAFILE:
		    {
			GHashTable *metadata;
			gboolean run_mf = TRUE;
			uint16_t ufid;
			gint ufid_gint;
			gchar *uri = NULL, *filename = NULL;

			/* Read in the mediafile */
			metadata = g_hash_table_new (g_str_hash, g_str_equal);

			while (run_mf)
			{
			    fread (&packet_id, sizeof(uint8_t), 1, fh);

			    switch (packet_id)
			    {
				case PACKET_MF_UFID:
				    {
					fread (&ufid, sizeof(uint16_t), 1, fh);
					break;
				    }

				case PACKET_MF_URI:
				    {
					fread (&value16, sizeof(uint16_t), 1, fh);

					uri = g_malloc0(value16+1);
					fread (uri, sizeof(char), value16, fh);
					break;
				    }

				case PACKET_MF_MVALUE:
				    {
					BmpMetadatumId tag_id;

					fread (&value, sizeof(uint8_t), 1, fh);
					tag_id = value;

					fread (&value16, sizeof(uint16_t), 1, fh);
					valuestr = g_malloc0(value16+1);
					fread (valuestr, sizeof(char), value16, fh);

					g_hash_table_insert (metadata, bmp_metadata_get_id (tag_id), valuestr);
					break;
				    }

				case PACKET_MF_EMF:
				    {
					run_mf = FALSE;
					break;
				    }

				default:
				    {
					g_message("%s: Unknown packet type %d while reading media file record", G_STRFUNC, packet_id);
				    }
			    }
			}

			if (! uri) g_message("%s: packet has no URI, discarding", G_STRFUNC);

			if (!g_ascii_strncasecmp("file",uri,4))
			  {
			    /* FIXME: Use GURI here */
			    filename =
				g_filename_from_uri (uri, NULL, NULL);

			    /* File is not present anymore, discard */
			    if (! g_file_test (filename, G_FILE_TEST_EXISTS)) {
				g_free (uri);
				g_free (filename);
				g_hash_table_destroy (metadata);
				break;
			    }
			    g_free (filename);
		          }

			g_hash_table_insert (metadata, bmp_metadata_get_id(BMP_DATUM_LOCATION), uri);
			g_hash_table_insert (metadata, "ufid", g_strdup_printf("%d", ufid));

			ufid_gint = ufid;

			if (self->private->ufid < ufid_gint) self->private->ufid = ufid_gint;

			g_hash_table_insert (self->private->ht_ufid, GINT_TO_POINTER(ufid_gint), metadata);
			g_hash_table_insert (self->private->ht_uri, g_strdup(uri), metadata);

			for (n = 0; n < BMP_DATUM_NONE; n++)
			{
			    GList *ufid_list = NULL;
			    gchar *key;

			    /* We acquire the particular metadata value from the just added file */
			    key = g_hash_table_lookup (metadata, bmp_metadata_get_id_static(n));

			    /* Now let's see if a list for this particular datum (Artist, Album...) already exists */
			    if (key)
			    {
				ufid_list = g_hash_table_lookup (self->private->ht_metadatum[n], key);

				if (! ufid_list) /*..nope*/ {
				    ufid_list = g_list_append (ufid_list, GINT_TO_POINTER(ufid_gint));
				    g_hash_table_insert (self->private->ht_metadatum[n], key, ufid_list);
				}
				else {
				    ufid_list = g_list_append (ufid_list, GINT_TO_POINTER(ufid_gint));
				    g_hash_table_insert (self->private->ht_metadatum[n], key, ufid_list);
				}
			    }
			}

		   	/* Now we insert into the specific artist/album hash tables for fast tree based overview creation
		         *
		         * This can be currently considered a hack, but since the relational stuff is in the pipeline, it doesn't
		         * really matter now.
		         */
			{
			    BmpUniqueList *_l_artist_albums = NULL, *_l_album_songs = NULL;
			    gchar *artist, *album;
			    gboolean _f_insert_artist = FALSE, _f_insert_album = FALSE;

			    artist = g_hash_table_lookup (metadata, bmp_metadata_get_id_static(BMP_DATUM_ARTIST));
			    album  = g_hash_table_lookup (metadata, bmp_metadata_get_id_static(BMP_DATUM_ALBUM));

			    if ((!artist) || (!strlen(artist)))
				artist = "(Unknown Artist)";
			    if ((!album) || (!strlen(album)))
				album = "(Unknown Album)";

			    /* Insert album for artist */
			    _l_artist_albums = g_hash_table_lookup (self->private->ht_artist_albums, artist);
			    _f_insert_artist = (_l_artist_albums != NULL) ? FALSE : TRUE;
			    if (_f_insert_artist)
				_l_artist_albums = bmp_unique_list_new(G_TYPE_STRING);
			    bmp_unique_list_insert (_l_artist_albums, g_strdup(album));
			    if (_f_insert_artist)
				g_hash_table_insert (self->private->ht_artist_albums, g_strdup(artist), _l_artist_albums);

			    /* Insert song for album */
			    _l_album_songs = g_hash_table_lookup (self->private->ht_album_songs, album);
			    _f_insert_album = (_l_album_songs != NULL) ? FALSE : TRUE;
			    if (_f_insert_album)
				_l_album_songs = bmp_unique_list_new (G_TYPE_INT);

			    bmp_unique_list_insert (_l_album_songs, GINT_TO_POINTER(ufid_gint));
			    if (_f_insert_album)
				g_hash_table_insert (self->private->ht_album_songs, g_strdup(album), _l_album_songs);
			}

		    }
		    break;

		    default:
		    {
			g_message("%s: Unknown packet type '%d' while reading metadata cache", G_STRFUNC, packet_id);
			break;
		    }
	    }
    }

    fclose (self->private->fh);
}

void
bmp_metadata_cache_validate (BmpMetadataCache *self)
{
    g_return_if_fail (BMP_IS_METADATA_CACHE (self));
}


/* Query API */

#define GET_VALUE(metadata,tag_id) g_hash_table_lookup(metadata,bmp_metadata_get_id_static(tag_id))
static void
query_process (gpointer _key, GHashTable *metadata, bmp_query_t *query)
{
    gboolean relevant = TRUE;
    GList *iter_list;

#if 0
    gchar *uri = NULL, *filename = NULL;

    uri =
	g_hash_table_lookup (metadata, bmp_metadata_get_id_static(BMP_DATUM_LOCATION));
    if (! uri)
	return;

    /* FIXME: Use GURI here */
    filename =
	g_filename_from_uri (uri, NULL, NULL);
    if (! filename)
	return;

    if (! g_file_test (filename, G_FILE_TEST_EXISTS)) {
	g_free (filename);
	return;
    }
    g_free (filename);
    g_free (uri);
#endif

    iter_list = query->query_atoms;

    while (iter_list && relevant)
      {
	bmp_query_atom_t *q_atom = iter_list->data;
	gchar *value, *cmp1, *cmp2;

	if (! q_atom->match_value)
	  {
	    relevant = TRUE;
	    break;
	  }

	value = GET_VALUE(metadata, q_atom->tag_id);

	if (! value)
          {
	    relevant = FALSE;
	    break;
	  }

	cmp1 = g_utf8_strdown (value, strlen(value));
	cmp2 = g_utf8_strdown (q_atom->match_value, strlen(q_atom->match_value));

	if (q_atom->exact_match)
          {
	    if (strcmp (cmp1, cmp2))
              {
		relevant = FALSE;
		g_free (cmp1);
		g_free (cmp2);
		break;
	      }
	    else
	      {
	        g_free (cmp1);
	        g_free (cmp2);
	      }
	  }
        else
          {
	    if (! g_strstr_len (cmp1, strlen(cmp1), cmp2))
              {
		relevant = FALSE;
		g_free (cmp1);
		g_free (cmp2);
		break;
	      }
	    else
	      {
	        g_free (cmp1);
	        g_free (cmp2);
	      }
	  }

	iter_list = g_list_next (iter_list);
      }

    if (relevant)
      {
	query->result_ufid_list = g_list_append (query->result_ufid_list, _key);
      }
}

static gint
query_sort (gconstpointer a, gconstpointer b, bmp_query_t *query) {

    BmpMetadataCache *self = BMP_METADATA_CACHE (query->user_data);
    GHashTable *metadata_1, *metadata_2;
    gchar *value1, *value2, *cmp1, *cmp2;
    gint result = 0, ia, ib;


    metadata_1 = g_hash_table_lookup (self->private->ht_ufid, a);
    metadata_2 = g_hash_table_lookup (self->private->ht_ufid, b);

    value1 = GET_VALUE (metadata_1, query->sort_tag_id);
    value2 = GET_VALUE (metadata_2, query->sort_tag_id);

    if ( (!value1) && (value2))  return -1;
    if ( (!value1) && (!value2)) return  0;
    if ( (value1)  && (!value2)) return  1;

    if ( (query->sort_tag_id == BMP_DATUM_TRACK) || (query->sort_tag_id == BMP_DATUM_DATE)) {

	ia = strtol (value1, NULL, 10);
	ib = strtol (value2, NULL, 10);

	result = ia-ib;

    } else {

	cmp1 = g_utf8_strdown (value1, strlen(value1));
	cmp2 = g_utf8_strdown (value2, strlen(value2));

	result = g_utf8_collate (cmp1, cmp2);

	g_free (cmp1);
	g_free (cmp2);
    }

    return result;
}

GList*
bmp_metadata_cache_query_extract (BmpMetadataCache *self, bmp_query_t *query, BmpMetadatumId tag_id)
{
    GList *result = NULL;
    GHashTable *ht_aux;

    for ( ; query->result_ufid_list; query->result_ufid_list = g_list_next (query->result_ufid_list))
      {
	 gchar *value;

         ht_aux =
	    bmp_metadata_cache_get_metadata_for_ufid (self, GPOINTER_TO_INT(query->result_ufid_list->data));

	 if (ht_aux)
           {
	      value =
	          g_hash_table_lookup (ht_aux, bmp_metadata_get_id_static (tag_id));

	      if (value)
                {
                   result = g_list_append (result, g_strdup(value));
                }

	   }
	 else
	   {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: System inconsistency: Metdata for UFID %d doesn't exist", G_STRFUNC, GPOINTER_TO_INT(query->result_ufid_list->data));
	   }

      }

   return result;
}

void
bmp_metadata_cache_query_run (BmpMetadataCache *self, bmp_query_t *query) {

    BmpMetadatumId sort_tag_id_save;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    /* Initialize result list */
    query->result_ufid_list = NULL;

    /* Set user_data to ourselves */
    query->user_data = self;

    g_hash_table_foreach (self->private->ht_ufid, (GHFunc)query_process, query);

    /* Sort */
    if (query->do_sort)
      {
	query->result_ufid_list = g_list_sort_with_data (query->result_ufid_list, (GCompareDataFunc)query_sort, query);
      }

    /* FIXME: Hack. We always pre-sort by track */
    sort_tag_id_save =
	    query->sort_tag_id;
    query->sort_tag_id =
	    BMP_DATUM_TRACK;
    query->result_ufid_list =
	    g_list_sort_with_data (query->result_ufid_list, (GCompareDataFunc)query_sort, query);
    query->sort_tag_id =
	    sort_tag_id_save;

}

bmp_query_t*
bmp_metadata_cache_query_construct (BmpMetadataCache *self, gboolean do_sort, BmpMetadatumId sort_tag_id, gint n_atoms, ...) {

/*
 *  EXAMPLE:
 *
 *	    bmp_query_t *query;
 *
 * 	    query = bmp_metadata_cache_query_construct (bmp_metadata_cache, TRUE, BMP_DATUM_TRACK,
 *							    2,
 *			    				    BMP_DATUM_ARTIST, "dr. dre", FALSE,
 *							    BMP_DATUM_ALBUM,  "2001", TRUE);
 *
 *  THIS EQUALS TO FOLLOWING SQL QUERY:
 *
 *	    SELECT ufid FROM metadata WHERE artist LIKE 'dr. dre' AND album='2001' ORDER BY track;
 *
 */

    bmp_query_t *query;
    va_list args;
    gint n;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), NULL);

    /* Initialize query */
    query = g_new0(bmp_query_t,1);
    query->do_sort = do_sort;
    query->sort_tag_id = sort_tag_id;
    query->query_atoms = NULL;

    va_start (args, n_atoms);

    for (n = 0; n < n_atoms; n++) {

	bmp_query_atom_t *query_atom = g_new0(bmp_query_atom_t,1);

	query_atom->tag_id =
	    va_arg (args, gint);
	query_atom->match_value =
	    g_strdup(va_arg (args, gchar*));
	query_atom->exact_match =
	    va_arg (args, gboolean);

	query->query_atoms = g_list_append (query->query_atoms, query_atom);
    }

    va_end (args);

    return query;
}

static void
dispose_query_atom (bmp_query_atom_t *atom, gpointer data) {

	g_free (atom->match_value);
	g_free (atom);
}

void
bmp_metadata_cache_query_dispose (BmpMetadataCache *self, bmp_query_t *query) {

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    g_list_foreach (g_list_first(query->query_atoms), (GFunc)dispose_query_atom, NULL);

    g_list_free (query->query_atoms);
    g_list_free (query->result_ufid_list);

    g_free (query);
}

#ifdef HAVE_GUI
static void
list_append_func (gpointer key, gpointer data, GtkListStore *store)
{
    GtkTreeIter iter;

    if (key && (strlen(key)>0))
      {
	if (g_utf8_validate(key, strlen(key), NULL))
          {
	    gtk_list_store_append (store, &iter);
	    gtk_list_store_set (store, &iter, 0, key, -1);
	  }
      }
}
#endif

static void
list_append_func_glist (gpointer key, gpointer data, GList **list)
{
    if (key && (strlen(key)>0))
      {
	if (g_utf8_validate(key, strlen(key), NULL))
          {
	    *list = g_list_append (*list, g_strdup(key));
	  }
      }
}

#ifdef HAVE_GUI
GtkListStore*
bmp_metadata_cache_list_tag_id (BmpMetadataCache *self, BmpMetadatumId tag_id)
{
    GtkListStore *result = gtk_list_store_new (1, G_TYPE_STRING);

    g_hash_table_foreach (self->private->ht_metadatum[tag_id],
                          (GHFunc) list_append_func, result);

    return result;
}
#endif

GList*
bmp_metadata_cache_list_tag_id_glist (BmpMetadataCache *self, BmpMetadatumId tag_id)
{
    GList *result = NULL;

    g_hash_table_foreach (self->private->ht_metadatum[tag_id],
                          (GHFunc) list_append_func_glist, &result);

    return result;
}

GList*
bmp_metadata_cache_list_tag_id_value (BmpMetadataCache *self, BmpMetadatumId tag_id, const gchar *value)
{
    return g_hash_table_lookup (self->private->ht_metadatum[tag_id], value);
}

gchar*
bmp_metadata_get_titlestring (GHashTable *metadata) {

    GString *string_out = NULL;
    GString *string_field = NULL;
    gchar *format, *format_save;
    gchar *field, *field_metadata;
    gint n;

    g_return_val_if_fail (metadata != NULL, NULL);

    LOOKUP_OR_CREATE_DEFAULT("bmp","titlestring-format","<artist> - <album> - <title>");
    format =
	bmp_configuration_key_fetch (bmp_configuration, "bmp", "titlestring-format");
    format_save =
	format;
    string_out =
	g_string_new ("");

    while (*format) {

	/* Begin sequence */
	if (*format == '<') {

	    string_field =
		g_string_new ("");
	    format++;

	    while (*format && (*format != '>') && (*format != ' ')) {
		g_string_append_c (string_field, *format);
		format++;
	    }

	    if (!*format)
		return g_string_free (string_out, FALSE);

	    field =
		g_string_free (string_field, FALSE);

	    /* Append metadata field */
	    for (n = 0; n < BMP_DATUM_NONE; n++) {
		if (! g_ascii_strcasecmp (field, bmp_metadata_get_id_static(n))) {
		    field_metadata =
			g_strdup (bmp_metadata_get_metadata_field (metadata, n));
		    if (field_metadata) {
			g_strstrip (field_metadata);
			g_string_append (string_out, field_metadata);
			g_free (field_metadata);
		    }
		    break;
		}
	    }
	} else {
	    g_string_append_c (string_out, *format);
	}
	format++;
    }

    g_free (format_save);

    return g_string_free (string_out, FALSE);
}


