#include <config.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#endif

#include <glib/gstdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <main.h>

#include "marshalers.h"
#include "g_object_get.h"

#include "bmpx/util.h"
#include "bmpx/vfs.h"
#include "bmpx/uri.h"
#include "bmpx/list.h"
#include "bmpx/plugin.h"
#include "bmpx/plugin_interfaces.h"
#include "bmpx/metadata.h"
#include "bmpx/playlist.h"

#include "plugin_priv.h"

G_DEFINE_TYPE (BmpPlaylist, bmp_playlist, G_TYPE_OBJECT)

typedef enum {
        BMP_PLAYLIST_SIGNAL_LIST_DELETED,
        BMP_PLAYLIST_SIGNAL_LIST_ADDED,
        BMP_PLAYLIST_SIGNAL_LIST_RENAMED,

        BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED,
	BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED,
	BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED,

	BMP_PLAYLIST_N_SIGNALS
} BmpPlaylistSignals;

static guint signals[BMP_PLAYLIST_N_SIGNALS] = { 0, 0, 0, 0, 0, 0};

typedef enum {
	BMP_PLAYLIST_PROP_0,

	BMP_PLAYLIST_N_PROPERTIES
} BmpPlaylistProperties;

struct _BmpPlaylistPrivate
{
        gboolean dispose_has_run;

	BmpListType *playlists;

	gchar  *pre_current_name;
	gint    pre_current_position;

	guint   guid_counter;
};

/* BMPx Playback History*/

#define PLAYBACK_HISTORY_LIMIT 1000

gboolean
row_guid_equal_func (RowGUID *guid_a, RowGUID *guid_b)
{
    g_return_val_if_fail (guid_a != NULL, FALSE);
    g_return_val_if_fail (guid_b != NULL, FALSE);

    if ((guid_a->counter == guid_b->counter) &&
        (guid_a->hash    == guid_b->hash))
    return TRUE;

    return FALSE;
}

guint
row_guid_hash_func (RowGUID *guid)
{
    guint  hash;
    gchar *hash_str;

    g_return_val_if_fail (guid != NULL, FALSE);

    hash_str = g_strdup_printf ("%u %u", guid->counter, guid->hash); 
    hash = g_str_hash (hash_str);
    g_free (hash_str);

    return hash;
}


static void
history_item_free (BmpPlaybackHistoryItem *item)
{
    g_return_if_fail (item != NULL);

    g_free (item->list_name);
    g_free (item->guid);
    g_free (item);
}

static void
history_free (BmpPlaybackHistory *history)
{
    gint n;

    g_return_if_fail (history != NULL);

    for (n = 0; n < history->length; n++)
      {
	 BmpPlaybackHistoryItem *item;

	 item = g_list_nth_data (history->history, n);
	 history_item_free (item);
      }

    g_list_free (history->history);
    g_free (history);
}

BmpPlaybackHistory*
bmp_playback_history_new (void)
{
    BmpPlaybackHistory* history;

    history = g_new0 (BmpPlaybackHistory,1);

    history->history = NULL;
    history->length  = -1;
    history->mark    = -1;

    return history;
}

void
bmp_playback_history_free (BmpPlaybackHistory *history)
{
    if (history == NULL)
	return;

    history_free (history);
}

static gboolean
history_item_prepend (BmpPlaybackHistory *history,
		      GObject *flow_plugin, 
		      gint tracklist)
{
    BmpPlaybackHistoryItem *item;

    g_return_val_if_fail (history != NULL, FALSE);

    item = bmp_plugin_flow_item_prev (flow_plugin, bmp_playlist_get_playlists(bmp_playlist), tracklist);

    if (!item)
	return FALSE;

    /* Did we reach the hard limit? If so, drop the head item */
    if (history->length+1 > PLAYBACK_HISTORY_LIMIT)
      {
	history_item_free (g_list_nth_data (history->history, g_list_length(history->history)-1));
	history->history = g_list_remove (history->history, g_list_nth_data (history->history,g_list_length(history->history)-1));
	history->length--;
      }

    history->history = g_list_prepend (history->history, item);
    history->length++;

    return TRUE;
}

static gboolean
history_item_append (BmpPlaybackHistory *history,
		     GObject *flow_plugin,
		     gint tracklist)
{
    BmpPlaybackHistoryItem *item;

    g_return_val_if_fail (history != NULL, FALSE);

    /* We check if we're not at the tail */
    if (history->mark < history->length)
      {
	g_list_nth (history->history, history->mark)->next = NULL;
	history->length = history->mark;
	/* FIXME: Free items */
      }

    item = bmp_plugin_flow_item_next (flow_plugin, bmp_playlist_get_playlists(bmp_playlist), tracklist);

    if (!item)
	return FALSE;

    /* Did we reach the hard limit? If so, drop the head item */
    if (history->length+1 > PLAYBACK_HISTORY_LIMIT)
      {
	history_item_free (g_list_nth_data (history->history, 0));
	history->history = g_list_remove (history->history, g_list_nth_data (history->history,0));
	history->length--;
      }

    history->history = g_list_append (history->history, item);
    history->length++;
    history->mark++;

    return TRUE;
}

gint
bmp_playback_history_rewind (BmpPlaybackHistory *history, gint tracklist)
{
    BmpPlaybackHistoryItem *history_item;
    gint row;

    g_return_val_if_fail (history != NULL, -1);

    history->mark = 0;

    history_item =
	g_list_nth_data (history->history, 0);
    row =
	bmp_playlist_get_row_by_guid (bmp_playlist, tracklist, history_item->guid);
    return row;
}

void
bmp_playback_history_item_set (BmpPlaybackHistory *history,
			       gint tracklist,
			       gint position)

{
    BmpPlaybackHistoryItem *item, *item_current = NULL;
    BmpListType *playlists;
    BmpPlaylistItem *playlist;
    BmpFileItem *file_item;

    g_return_if_fail (history != NULL);


    /* We check if we're not at the tail */
    if (history->mark < history->length)
      {
	g_list_nth (history->history, history->mark)->next = NULL;
	history->length = history->mark;
	/* FIXME: Free items */
      }

    playlists =
	(BmpListType *)bmp_playlist_get_playlists (bmp_playlist);
    playlist =
	g_list_nth_data(playlists->items, tracklist);

    item = g_new0 (BmpPlaybackHistoryItem,1);

    item->list_name = g_strdup (playlist->listname);
    item->list_time = playlist->list->time;

    file_item =
	BMP_FILE_ITEM(g_list_nth_data(playlist->list->items, position));

    if (history->length > 0)
	item_current = g_list_nth_data (history->history, history->mark);

    if (item_current)
      {
	RowGUID *guid_a, *guid_b;

	guid_a = item_current->guid;
	guid_b = file_item->guid;

	if (row_guid_equal_func (guid_a, guid_b))
	  {
	    history_item_free (item);
	    return;
	  }
      }

    item->guid = g_malloc0(sizeof(RowGUID));
    memcpy (item->guid, file_item->guid, sizeof(RowGUID));

    /* Did we reach the hard limit? If so, drop the head item */
    if (history->length+1 > PLAYBACK_HISTORY_LIMIT)
      {
	history_item_free (g_list_nth_data(history->history, 0));
	history->history = g_list_remove (history->history, g_list_nth_data (history->history, 0));
	history->length--;
      }
    history->history = g_list_append (history->history, item);
    history->length++;
    history->mark++;
}

gint
bmp_playback_history_item_back (BmpPlaybackHistory *history, GObject *flow_plugin, gint tracklist)
{
    BmpPlaybackHistoryItem *history_item;
    gint row;

    g_return_val_if_fail (history != NULL, -1);

    if (!flow_plugin)
	return -1;

    item_back_iteration:

    if (history->mark == 0)
      {
	if (history_item_prepend (history, flow_plugin, tracklist))
	  {
	    history_item =
		g_list_nth_data (history->history, history->mark);
	    row =
		bmp_playlist_get_row_by_guid (bmp_playlist, tracklist, history_item->guid);
	    return row;
	  }
	else
	  {
	    return -1;
	  }
      }
    else
      {
	history->mark--;
	history_item =
	    g_list_nth_data (history->history, history->mark);
	row =
	    bmp_playlist_get_row_by_guid (bmp_playlist, tracklist, history_item->guid);
	if (row == -1)
	  {
	    g_list_remove (history->history, history_item);
	    history_item_free (history_item);
	    goto item_back_iteration;
	  }
      }

    return row;
}

gint
bmp_playback_history_item_forward (BmpPlaybackHistory *history, GObject *flow_plugin, gint tracklist)
{
    BmpPlaybackHistoryItem *history_item;
    gint row;

    g_return_val_if_fail (history != NULL, -1);

    if (!flow_plugin)
	return -1;

    item_forward_iteration:

    if (history->mark == history->length)
      {
	if (history_item_append (history, flow_plugin, tracklist))
	  {
	    history_item =
		g_list_nth_data (history->history, history->mark);
	    row =
		bmp_playlist_get_row_by_guid (bmp_playlist, tracklist, history_item->guid);
	    return row;
	  }
	else
	  {
	    return -1;
	  }
      }
    else
      {
	history->mark++;

	history_item =
	    g_list_nth_data (history->history, history->mark);
	row =
	    bmp_playlist_get_row_by_guid (bmp_playlist, tracklist, history_item->guid);
	if (row == -1)
	  {
	    g_list_remove (history->history, history_item);
	    history_item_free (history_item);
	    goto item_forward_iteration;
	  }
      }

    return row;
}


/* Track List Funcs */
gchar*
tracklist_column_data_func (BmpFileItem *item, gint column_idx)
{
    gchar *response = NULL;

#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	switch (column_idx)
	{
		case 0:
		{
			if (item->metadata) {
				response = bmp_metadata_get_titlestring (item->metadata);
			} else {
				response = g_path_get_basename (item->uri);
			}
		}
		break;

		case 1:
		{

			gint ttime;
			gint minutes;
			gint seconds;
			const gchar *time;

			if (!item->metadata)
                            return g_strdup("");

			time = bmp_metadata_get_metadata_field(item->metadata, BMP_DATUM_TIME);

			if (time) {
			    ttime = strtol (time,NULL,10);
			}
			else
			    ttime = 0;

			minutes = ttime / 60;
			seconds = ttime % 60;

			response = g_strdup_printf("%d:%-2.2d",minutes,seconds);
		}
		break;
	}

	return response;
}

static void
tracklist_item_swap_func (GList *items, gint _idx1, gint _idx2)
{
	BmpFileItem *fi_a, *fi_b;

 	GList *a, *b;

	a = g_list_nth (items, _idx1);
	b = g_list_nth (items, _idx2);

	glist_swap(a,b);

	fi_a = a->data;
	fi_b = b->data;

	fi_a->dirty = TRUE;
	fi_b->dirty = TRUE;
}


void
tracklist_on_display_func (BmpFileItem *item)
{
	if (!item) return;
	if (item->metadata_attempted) return;
	if (item->metadata != NULL) return;

	item->metadata = bmp_metadata_cache_get_metadata_for_uri (bmp_metadata_cache, item->uri);
	item->metadata_attempted = TRUE;
	item->dirty = TRUE;
}

static gboolean
tracklist_get_dirty_func (BmpFileItem *item)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return FALSE;
	return item->dirty;
}

static void
tracklist_set_dirty_func (BmpFileItem *item, gboolean dirty)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return;
	item->dirty = dirty;
}

static gboolean
tracklist_get_rendered_func (BmpFileItem *item)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return FALSE;
	return item->rendered;
}

static void
tracklist_set_rendered_func (BmpFileItem *item, gboolean rendered)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	if (!item) return;
	item->rendered = rendered;
}


static gboolean
tracklist_get_selected_func (BmpFileItem *item)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	return item->selected;
}

static void
tracklist_set_selected_func (BmpFileItem *item, gboolean selected)
{
#if 0
	while (item->locked) { while (gtk_events_pending()) gtk_main_iteration(); }
#endif

	item->selected = selected;
	item->dirty = TRUE;
}

static void
tracklist_item_destroy_func (BmpFileItem *item)
{
    g_free (item->uri);
    g_free (item);
}

/* Playlists List funcs */
static gchar*
playlists_column_data_func (BmpPlaylistItem *item, gint column_idx)
{
	if (!item) return NULL;

	return item->listname;
}

static gboolean
playlists_get_dirty_func (BmpPlaylistItem *item)
{
	if (!item) return FALSE;
	return item->dirty;
}

static void
playlists_set_dirty_func (BmpPlaylistItem *item, gboolean dirty)
{
	if (!item) return;
	item->dirty = dirty;
}

static gboolean
playlists_get_rendered_func (BmpPlaylistItem *item)
{
	if (!item) return FALSE;
	return item->dirty;
}

static void
playlists_set_rendered_func (BmpPlaylistItem *item, gboolean dirty)
{
	if (!item) return;
	item->dirty = dirty;
}


static gboolean
playlists_get_selected_func (BmpPlaylistItem *item)
{
	if (!item) return FALSE;

	return item->selected;
}

static void
playlists_set_selected_func (BmpPlaylistItem *item, gboolean selected)
{
	item->selected = selected;
	item->dirty = TRUE;
}

static void
playlists_item_destroy_func (BmpPlaylistItem *item)
{
	/*g_free (item->filename);
	g_free (item->listname);*/
}

static gint
playlists_compare_func (const BmpPlaylistItem *a, const BmpPlaylistItem *b)
{
	gchar *string_a;
	gchar *string_b;
	gint cmp_result;

	string_a = g_utf8_casefold (a->listname, strlen(a->listname));
	string_b = g_utf8_casefold (b->listname, strlen(b->listname));

	if ( (string_a != NULL) && (string_b == NULL)) return  1;
	if ( (string_a == NULL) && (string_b == NULL)) return  0;
	if ( (string_a == NULL) && (string_b != NULL)) return -1;

	cmp_result = g_utf8_collate (string_a, string_b);

	/*g_free (string_a);
	g_free (string_b);*/

	return cmp_result;
}

static gint
tracklist_compare_func (BmpFileItem *a, BmpFileItem *b, gpointer data)
{
	BmpMetadatumId tag_id;
	gchar *cmptext_a, *cmptext_b;
	gchar *string_a;
	gchar *string_b;
	gint cmp_result;
        gint val_a, val_b;

	tracklist_on_display_func (a);
	tracklist_on_display_func (b);

	if (! data)
	{
		cmptext_a = tracklist_column_data_func (a, 0);
		cmptext_b = tracklist_column_data_func (b, 0);

		if ( (cmptext_a != NULL) && (cmptext_b == NULL)) return  1;
		if ( (cmptext_a == NULL) && (cmptext_b == NULL)) return  0;
		if ( (cmptext_a == NULL) && (cmptext_b != NULL)) return -1;

	}
	else
	{
		tag_id = GPOINTER_TO_INT(data);

		cmptext_a = g_hash_table_lookup(a->metadata, bmp_metadata_get_id_static(tag_id));
		cmptext_b = g_hash_table_lookup(b->metadata, bmp_metadata_get_id_static(tag_id));

		if ( (cmptext_a != NULL) && (cmptext_b == NULL)) return  1;
		if ( (cmptext_a == NULL) && (cmptext_b == NULL)) return  0;
		if ( (cmptext_a == NULL) && (cmptext_b != NULL)) return -1;

		if ((tag_id == BMP_DATUM_TRACK) || (tag_id == BMP_DATUM_DATE))
		{
			if (! cmptext_a)
			    val_a = 0;
			else
			    val_a = strtol(cmptext_a,NULL,10);

			if (! cmptext_b)
			    val_b = 0;
			else
			    val_b = strtol(cmptext_b,NULL,10);

			return val_a - val_b;
		}
	}

	string_a = g_utf8_casefold (cmptext_a, strlen(cmptext_a));
	string_b = g_utf8_casefold (cmptext_b, strlen(cmptext_b));

	if ( (string_a != NULL) && (string_b == NULL)) return  1;
	if ( (string_a == NULL) && (string_b == NULL)) return  0;
	if ( (string_a == NULL) && (string_b != NULL)) return -1;

	cmp_result = g_utf8_collate (string_a, string_b);

	g_free (string_a);
	g_free (string_b);

	return cmp_result;
}


#if 0
static void
_print_listnames (BmpPlaylistItem *item, gpointer data)
{
	g_message ("LIST: %s", item->listname);
}
#endif

static void
playlists_sort_func (BmpPlaylist *self, gconstpointer data2)
{
	self->private->playlists->items =
			g_list_sort (self->private->playlists->items,
			(GCompareFunc)playlists_compare_func);

}

static void
tracklist_sort_func (BmpListType *list, gconstpointer data2)
{
	gpointer item;

	item = g_list_nth (list->items, list->current);
	list->items = g_list_sort_with_data (list->items, (GCompareDataFunc)tracklist_compare_func, (gpointer)data2);
	list->current = g_list_position (list->items, item);
}

static void
_playlists_clear_current (gpointer item, BmpPlaylist *self)
{
	BmpPlaylistItem *list_item = (BmpPlaylistItem*)item;

	list_item->list->current = -1;
}

void
bmp_playlist_sort_current_list (BmpPlaylist *self)
{
	BmpListType *list;
	GList *llink = NULL;
	gint _idx;

	_idx =
	    g_object_get_int(bmp_system_control, "current_tracklist");
	list =
	    bmp_playlist_tracklist_get_i (self, _idx);

	if (list->last_selected != -1)
	    llink =
		g_list_nth (list->items, list->last_selected);

	list->list_sort_func (list, NULL);

	if (llink)
	    list->last_selected =
		g_list_position (list->items, llink);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED],
                   0, _idx);
}

void
bmp_playlist_sort_current_list_by_tag (BmpPlaylist *self, BmpMetadatumId tag_id)
{
	BmpListType *list;
	GList *llink = NULL;
	gint _idx;

	_idx =
	    g_object_get_int(bmp_system_control, "current_tracklist");
	list =
	    bmp_playlist_tracklist_get_i (self, _idx);

	if (list->last_selected != -1)
	    llink =
		g_list_nth (list->items, list->last_selected);

	list->list_sort_func (list, GINT_TO_POINTER(tag_id));

	if (llink)
	    list->last_selected =
		g_list_position (list->items, llink);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED],
                   0, _idx);
}


void
bmp_playlist_clear_all_current (BmpPlaylist *self)
{
	g_return_if_fail (BMP_IS_PLAYLIST (self));

	g_list_foreach (self->private->playlists->items, (GFunc)_playlists_clear_current, self);
}

static void
load_playlist (BmpPlaylist *self, const gchar *filename)
{

	gchar *name, *aux;
	gchar **uri_strv;
	gint new_count = 0;
	gint list;
	GURI *g_uri;
	VFSHandle *handle;
	NGC *gc_local;

	g_return_if_fail (BMP_IS_PLAYLIST (self));

	gc_local = gc_new();
	gc_mark (gc_local);

	aux =
	    gc_collect (gc_local, g_path_get_basename (filename));
	name =
	    g_strndup (aux, strlen(aux) - 4);
	list =
	    bmp_playlist_tracklist_new (self, name);
	aux =
	    gc_collect (gc_local, g_strconcat ("file://", filename, NULL));
	g_uri =
	    bmp_uri_new (aux);
	bmp_uri_escape (g_uri);
	handle =
	    vfs_handle_new (gc_collect(gc_local,bmp_uri_get_string(g_uri)));

	vfs_handle_read_uris (handle);

	uri_strv =
	    glist_to_strv (handle->uri_list);


	bmp_system_control_add_uri_list (bmp_system_control, uri_strv, list, -1, &new_count, NULL);
	vfs_handle_free (handle);
        g_signal_emit (G_OBJECT (self), signals[BMP_PLAYLIST_SIGNAL_LIST_ADDED], 0, list);


	gc_free (gc_local);
	gc_destroy (gc_local);

#if 0
	if (!strcmp(name, self->private->pre_current_name) && (self->private->pre_current_position != -1)) {

		list->current = self->private->pre_current_position;
		self->private->pre_current_position = -1;
	}
	else
		list->current = -1;
#endif

	g_free(name);
}

#if 0
static void
playlist_removed_func (const gchar *filename, BmpPlaylist *self)
{
}

static void
playlist_added_func (const gchar *filename, BmpPlaylist *self)
{
	gchar *fqname;

	fqname = g_build_filename (bmp_configuration_key_fetch (bmp_configuration, "bmp", "playlists-path"),
				   filename, NULL);
	load_playlist (self, fqname);
	g_free (fqname);

}
#endif

static void
bmp_playlist_init (BmpPlaylist  *self)
{
	gchar *listname;
	gint _idx;

        self->private = g_new0(BmpPlaylistPrivate, 1);

        self->private->dispose_has_run = FALSE;

	self->private->pre_current_position = -1;
	self->private->pre_current_name = NULL;

	self->private->playlists = g_new0(BmpListType,1);
	self->private->playlists->name = NULL;
	self->private->playlists->column_data_func = (ColumnDataFunc)playlists_column_data_func;
	self->private->playlists->row_get_selected_func = (RowGetSelectedFunc)playlists_get_selected_func;
	self->private->playlists->row_set_selected_func = (RowSetSelectedFunc)playlists_set_selected_func;
	self->private->playlists->list_sort_func = (ListSortFunc)playlists_sort_func;
	self->private->playlists->on_display_func = NULL;
	self->private->playlists->get_dirty_func = (GetDirtyFunc)playlists_get_dirty_func;
	self->private->playlists->set_dirty_func = (SetDirtyFunc)playlists_set_dirty_func;
	self->private->playlists->item_destroy_func = (GDestroyNotify)playlists_item_destroy_func;
	self->private->playlists->set_rendered_func = (SetRenderedFunc)playlists_set_rendered_func;
	self->private->playlists->get_rendered_func = (GetRenderedFunc)playlists_get_rendered_func;
	self->private->playlists->items = NULL;
	self->private->playlists->current = -1;
	self->private->playlists->last_selected = -1;
	self->private->playlists->multiple_selected = FALSE;

	self->private->guid_counter = 0;

#if 0
	fam_monitor_watch_new (bmp_fam_monitor, "playlists",
				bmp_configuration_key_fetch(bmp_configuration, "bmp", "playlists-path"),
				playlist_added_func,
				playlist_removed_func,
				self,
				NULL);
#endif

    listname =
	bmp_configuration_key_fetch(bmp_configuration,"bmp","current-playlist");
    _idx =
	bmp_configuration_key_fetch_int(bmp_configuration,"bmp","current-playlist-position");

    if (! listname) listname = g_strdup("");

    bmp_playlist_set_pre_current_name (self, listname);
    bmp_playlist_set_pre_current_position (self, _idx);

    g_free (listname);

    g_message("%s\t\t[INITIALIZED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_playlist_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static void
bmp_playlist_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static GObject *
bmp_playlist_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpPlaylistClass *klass;
                GObjectClass *parent_class;
                klass = BMP_PLAYLIST_CLASS (g_type_class_peek (BMP_TYPE_PLAYLIST));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
bmp_playlist_dispose (GObject *obj)
{
    BmpPlaylist *self = (BmpPlaylist *) obj;

    if (self->private->dispose_has_run)
        return;

    self->private->dispose_has_run = TRUE;

    bmp_playlist_save_lists (self);

    g_message("%s\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(obj));
}

static void
bmp_playlist_finalize (GObject *object)
{
    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */

#if 0
    BmpPlaylist *self = (BmpPlaylist *) gobject;
    g_free (self->private);
#endif
}

BmpPlaylist*
bmp_playlist_new (void)
{
  BmpPlaylist *list;
  list = g_object_new (bmp_playlist_get_type (), NULL);

  return list;
}

static void
bmp_playlist_class_init (BmpPlaylistClass *g_class)

{
    GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

    gobject_class->set_property = bmp_playlist_set_property;
    gobject_class->get_property = bmp_playlist_get_property;
    gobject_class->dispose = bmp_playlist_dispose;
    gobject_class->finalize = bmp_playlist_finalize;
    gobject_class->constructor = bmp_playlist_constructor;

    signals[BMP_PLAYLIST_SIGNAL_LIST_DELETED] =
        g_signal_new ("list-deleted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, list_deleted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_LIST_ADDED] =
        g_signal_new ("list-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, list_added),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_LIST_RENAMED] =
        g_signal_new ("list-renamed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, list_renamed),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_LIST_SORTED] =
        g_signal_new ("tracklist-list-sorted",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_list_sorted),
                      NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE,
                      1, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
        g_signal_new ("tracklist-items-removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_removed),
                      NULL, NULL, bmpx_VOID__INT_POINTER_INT, G_TYPE_NONE,
                      3, G_TYPE_INT, G_TYPE_POINTER, G_TYPE_INT);

    signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED] =
        g_signal_new ("tracklist-items-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (BmpPlaylistClass, tracklist_items_added),
                      NULL, NULL, bmpx_VOID__INT_POINTER_INT, G_TYPE_NONE,
                      3, G_TYPE_INT, G_TYPE_POINTER, G_TYPE_INT);
}

const BmpListType*
bmp_playlist_get_playlists (BmpPlaylist *self)
{
	return self->private->playlists;
}

const gchar*
bmp_playlist_get_nth_tracklist_name (BmpPlaylist *self, gint _idx)
{
	BmpPlaylistItem *item;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), NULL);

	if (! self->private->playlists->items)
	    return NULL;
	if (g_list_length (self->private->playlists->items) == 0)
	    return NULL;
	if (_idx > (g_list_length (self->private->playlists->items)-1))
	    return NULL;

	item = LIST_GET_NTH_ITEM(self->private->playlists, _idx);

	return item->listname;
}

void
bmp_playlist_set_pre_current_position (BmpPlaylist *self, gint _idx)
{
	g_return_if_fail (BMP_IS_PLAYLIST (self));

	self->private->pre_current_position = _idx;
}

void
bmp_playlist_set_pre_current_name (BmpPlaylist *self, const gchar *name)
{
	g_return_if_fail (BMP_IS_PLAYLIST (self));

	self->private->pre_current_name = g_strdup (name);
}

const gchar*
bmp_playlist_get_current_tracklist_name (BmpPlaylist *self)
{
	BmpPlaylistItem *item;
	gint _idx;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), NULL);

	if (! self->private->playlists)
	    return NULL;
	if (! self->private->playlists->items)
	    return NULL;
	if (g_list_length(self->private->playlists->items) == 0)
	    return NULL;

	_idx = g_object_get_int (bmp_system_control, "current_tracklist");

	if (_idx < 0)
	    return NULL;

	item = LIST_GET_NTH_ITEM(self->private->playlists, _idx);

	return item->listname;
}

gint
bmp_playlist_get_tracklist_by_name (BmpPlaylist *self, const gchar *name)
{
	BmpPlaylistItem *item;
	GList *iter_list;
	gint ctr = 0;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), -1);

	iter_list = self->private->playlists->items;

	while (iter_list)
	{
		item = iter_list->data;
		if (!strcmp(name, item->listname)) break;
		ctr++;

		iter_list = g_list_next (iter_list);
	}

	return ctr;
}

static gboolean
playlist_rename (BmpPlaylist *self, const gchar *listname, const gchar *newname)
{
	BmpListType *tracklist;
	BmpPlaylistItem *playlist_item;
	GList *iter_list;
	gchar *filename, *aux1;

	g_return_val_if_fail (BMP_IS_PLAYLIST (self), FALSE);

	tracklist = bmp_playlist_tracklist_get (self, newname);
	if (tracklist) return FALSE;

	iter_list = g_list_first(self->private->playlists->items);

	while (iter_list)
	{
		playlist_item = iter_list->data;

		if (!strcmp(playlist_item->listname, listname))
		{
			aux1 = g_strconcat (newname, ".m3u", NULL);
			filename = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], aux1, NULL);

			g_rename (playlist_item->filename, filename);

			g_free (aux1);
			g_free (filename);
			break;
		}
		iter_list = g_list_next (iter_list);
	}

	return TRUE;
}

gint
bmp_playlist_tracklist_rename (BmpPlaylist *self, gint _idx, const gchar *name)
{
	BmpPlaylistItem *item;
	GList *llink_current = NULL, *llink_active = NULL, *llink = NULL, *llink_save;
	gchar *name_old, *aux1, *aux2;

	llink =
	    g_list_nth (self->private->playlists->items, self->private->playlists->current);
	llink_save =
	    g_list_nth (self->private->playlists->items, _idx);
	llink_current =
	    g_list_nth(self->private->playlists->items, g_object_get_int (bmp_system_control, "current_tracklist"));
	llink_active  =
	    g_list_nth(self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));

	item = LIST_GET_NTH_ITEM(self->private->playlists, _idx);
	name_old = g_strdup (item->listname);
	if (! playlist_rename (self, name_old, name)) return -1;

	aux1 = g_strconcat(name, ".m3u", NULL);
	aux2 = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], aux1, NULL);
	g_free (item->listname);
	item->listname = g_strdup(name);
	g_free (item->filename);
	item->filename = g_strdup(aux2);

	self->private->playlists->list_sort_func (self, NULL);

	self->private->playlists->current = g_list_position(self->private->playlists->items, llink);
	g_object_set (bmp_system_control, "current_tracklist", (gint)g_list_position(self->private->playlists->items, (GList*)llink_current), NULL);
	g_object_set (bmp_system_control, "active_tracklist",  (gint)g_list_position(self->private->playlists->items, (GList*)llink_active), NULL);

	g_free (aux1);
	g_free (aux2);
	g_free (name_old);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_RENAMED],
                   0, _idx);

	return g_list_position (self->private->playlists->items, llink_save);
}

void
bmp_playlist_tracklists_remove_all (BmpPlaylist *self)
{
	BmpPlaylistItem *item;
	GList *iter_list;

	iter_list = self->private->playlists->items;

	while (iter_list)
	{
		item = iter_list->data;
		g_unlink (item->filename);
		iter_list = g_list_next (iter_list);
	}

	self->private->playlists->current = -1;
	self->private->playlists->items = NULL;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_DELETED],
                   0, -1);
}

gboolean
bmp_playlist_tracklist_remove (BmpPlaylist *self, gint _idx)
{
	BmpPlaylistItem *item;
	GList *llink_current = NULL, *llink_active = NULL, *llink = NULL;

	item = g_list_nth(self->private->playlists->items, _idx)->data;

	if (!item)
	    return FALSE;

	g_unlink (item->filename);

	llink_current = g_list_nth(self->private->playlists->items, g_object_get_int (bmp_system_control, "current_tracklist"));
	llink_active  = g_list_nth(self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	llink = g_list_nth (self->private->playlists->items, self->private->playlists->current);

	LIST_REMOVE_ITEM(self->private->playlists, _idx);

	/* FIXME: We emit the signal here so the handlers still have the correct current_tracklist to compare
	          against, but is it really ok to do it here and not at the end of the function */
        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_DELETED],
                   0, _idx);

	self->private->playlists->current = g_list_position (self->private->playlists->items, llink);
	g_object_set (bmp_system_control, "current_tracklist", (gint)g_list_position(self->private->playlists->items, llink_current), NULL);
	g_object_set (bmp_system_control, "active_tracklist",  (gint)g_list_position(self->private->playlists->items, llink_active), NULL);

	return TRUE;
}

#if 0
static void
playlist_tracklist_remove_by_name (BmpPlaylist *self, const gchar *name)
{
}
#endif

static gint
insert_playlist_item (BmpPlaylist *self, const gchar *name, BmpListType *list)
{
	BmpPlaylistItem *playlist_item = NULL;
	GList *llink;
	gchar *basename;
	gchar *filename;

	llink = g_list_nth (self->private->playlists->items, self->private->playlists->current);

	basename = g_strconcat (name, ".m3u", NULL);
	filename = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], basename, NULL);
	g_free (basename);

	playlist_item = g_new0(BmpPlaylistItem,1);
	playlist_item->filename = filename;
	playlist_item->listname = g_strdup(name);
	playlist_item->selected = FALSE;
	playlist_item->dirty = FALSE;
	playlist_item->rendered = FALSE;
	playlist_item->list = list;
	LIST_ADD_ITEM(self->private->playlists, playlist_item);

	self->private->playlists->list_sort_func (self, NULL);
	self->private->playlists->current = g_list_position (self->private->playlists->items, llink);

	return g_list_position (self->private->playlists->items, g_list_find (self->private->playlists->items, playlist_item));
}

static BmpFileItem*
file_item_new (BmpPlaylist *self, const gchar *uri) {

    BmpFileItem *file_item = NULL;
    RowGUID *row_guid;

    g_return_val_if_fail (uri != NULL, NULL);

    file_item = g_new0(BmpFileItem,1);
    row_guid = g_new0(RowGUID,1);

    row_guid->counter =
	self->private->guid_counter++;
    row_guid->hash =
	g_str_hash(uri);

    file_item->uri = g_strdup(uri);
    file_item->metadata = NULL;
    file_item->metadata_attempted = FALSE;
    file_item->selected = FALSE;
    file_item->dirty = FALSE;
    file_item->rendered = FALSE;
    file_item->guid = row_guid;

    return file_item;
}

gint
bmp_playlist_get_row_by_guid (BmpPlaylist *self, gint tracklist, RowGUID *guid)
{
    BmpListType *list;
    BmpFileItem *item;
    gint n = 0;
    gboolean found = FALSE;
    GList *iter_list;

    g_return_val_if_fail (BMP_IS_PLAYLIST (self), -1);

    list =
	bmp_playlist_tracklist_get_i (self, tracklist);
    iter_list = list->items;


    for ( ; iter_list; iter_list = iter_list->next)
      {
	RowGUID *guid_cmp;

	item =
	    BMP_FILE_ITEM(iter_list->data);
	guid_cmp =
	    item->guid;

	if ((guid->counter == guid_cmp->counter) && (guid->hash == guid_cmp->hash))
	 {
	    found = TRUE;
	    break;
	 }
	n++;
      }

    return found?n:-1;
}

static int
bmp_playlist_insert_tracklist_item_position_R (BmpPlaylist *self, gint tracklist, const gchar *_uri, gint _idx)
{
    BmpFileItem *file_item = NULL;
    GURI *g_uri;
    VFSHandle *handle;
    gchar *uri;

    g_uri =
	bmp_uri_new (_uri);
    if (! g_uri)
	return _idx;
    bmp_uri_unescape (g_uri);

    if (G_URI_IS_LOCAL(_uri) && g_file_test (g_uri->path, G_FILE_TEST_IS_REGULAR) && is_audio_file(g_uri->path))
      {
	file_item = file_item_new (self, _uri);
	LIST_ADD_ITEM_POSITION(bmp_playlist_tracklist_get_i (self, tracklist), file_item, _idx++);
      }
    else
      {
	uri =
	    bmp_uri_get_string (g_uri);
        handle =
	    vfs_handle_new (_uri);
	if (!vfs_handle_read_uris (handle))
	  {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Unable to process URI '%s'", G_STRLOC, _uri);
	    vfs_handle_free (handle);
	    bmp_uri_delete (g_uri);
	    return _idx;
          }

	 for (; handle->uri_list; handle->uri_list = handle->uri_list->next)
	   {
	      file_item = file_item_new (self, handle->uri_list->data);
	      LIST_ADD_ITEM_POSITION(bmp_playlist_tracklist_get_i (self, tracklist), file_item, _idx++);
	   }

	vfs_handle_free (handle);
	g_free (uri);
      }

    bmp_uri_delete (g_uri);
    return _idx;
}

gint
bmp_playlist_insert_tracklist_items_position (BmpPlaylist *self, gint tracklist, const gchar **uri_list, gint _idx)
{
    gint n = 0, n_added_rows = 0, _idx_original = _idx;
    GArray *added_rows;

    while (uri_list[n])
      {
        _idx = bmp_playlist_insert_tracklist_item_position_R (self, tracklist, uri_list[n], _idx);
	n++;
      }

    added_rows = g_array_new (FALSE, FALSE, sizeof(gint));

    for (n = 0; n < (_idx - _idx_original); n++)
      {
	gint val =
	    n + _idx_original;
	g_array_append_val (added_rows, val);
      }

    n_added_rows = (_idx - _idx_original);

    g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_ADDED],
                   0, tracklist, added_rows, n_added_rows);

    return _idx;
}

gint
bmp_playlist_insert_tracklist_items (BmpPlaylist *self, gint tracklist, const gchar **uri_list)
{
    BmpListType *list;
    gint _idx;

    list = bmp_playlist_tracklist_get_i (self, tracklist);

    if (!list)
    {
        g_message ("%s: NO LIST! returning.", G_STRLOC);
        return -1;
    }

    _idx =
	g_list_length (list->items);
    _idx =
	bmp_playlist_insert_tracklist_items_position (self, tracklist, uri_list, _idx);

   return _idx;
}

void
bmp_playlist_tracklist_remove_rows_all (BmpPlaylist *self, gint tracklist)
{
	BmpListType *list;
	GArray *removed_rows;
	gint n_removed_rows = -1;

	removed_rows = g_array_new (FALSE,FALSE,sizeof(gint));

	list = bmp_playlist_tracklist_get_i (self, tracklist);
	g_list_foreach (list->items, (GFunc) tracklist_item_destroy_func, NULL);
	list->items = NULL;
	list->current = -1;
	list->last_selected = LAST_SELECTED_NONE;
	list->adjustment = 0.0;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, tracklist, removed_rows, n_removed_rows);

	/* NULL means: All. heh. */
}

void
bmp_playlist_tracklist_remove_selected (BmpPlaylist *self, gint tracklist)
{
	GList *new_items;
	BmpListType *list;
	BmpFileItem *file_item;
	gint n, new_current, new_count, n_removed_rows = 0;
	GArray *removed_rows;

	list =
	    bmp_playlist_tracklist_get_i (self, tracklist);
	new_items =
	    NULL;
	new_count =
	    0;
	new_current =
	    -1;
	removed_rows =
	    g_array_new (FALSE, FALSE, sizeof(gint));

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		file_item = LIST_GET_NTH_ITEM (list, n);

		/* FIXME: is it really ok to access privately? by doing this be basically say, the function pointers
			  in the list struct are only for the chromalist widget.. */
		if (!file_item->selected)
		  {
			new_items = g_list_append (new_items, file_item);
			if (n == list->current)
				new_current = new_count;
			new_count++;
		  }
		else
                  {
		    g_array_append_val (removed_rows, n);
		    n_removed_rows++;
		  }
	   }
	list->items = new_items;
	list->current = new_current;

	if ((list->last_selected > g_list_length(list->items)-1) && new_items)
	    list->last_selected = g_list_length(list->items)-1;
	else if (! new_items)
	    list->last_selected = LAST_SELECTED_NONE;

	if (list->last_selected != LAST_SELECTED_NONE)
	{
	    file_item = g_list_nth(list->items, list->last_selected)->data;
	    list->row_set_selected_func ((gpointer)file_item, TRUE);
	}
	else
	{
	    list->last_selected = LAST_SELECTED_NONE;
	}

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, tracklist, removed_rows, n_removed_rows);

	/* FIXME: Free old items */
}

void
bmp_playlist_tracklist_remove_rows (BmpPlaylist *self, gint tracklist, GArray *rows, gint n_rows)
{
	GList *new_items;
	BmpListType *list;
	BmpFileItem *file_item;
	GHashTable *mapping;
	gint n, new_count, new_current;

	list =
	    bmp_playlist_tracklist_get_i (self, tracklist);
	new_items =
	    NULL;
	new_count =
	    0;
	new_current =
	    -1;
	mapping = g_hash_table_new (g_direct_hash, g_direct_equal);

	for (n = 0; n < n_rows; n++)
	  {
	    gint val;
	    val =
		g_array_index (rows, gint, n);
	    g_hash_table_insert (mapping, GINT_TO_POINTER(val),
				 GINT_TO_POINTER(0x01));
	  }

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		file_item = LIST_GET_NTH_ITEM (list, n);

		if (! g_hash_table_lookup (mapping, GINT_TO_POINTER(n)))
		  {
			new_items = g_list_append (new_items, file_item);
			if (n == list->current)
				new_current = new_count;
			new_count++;
		  }
	  }

	list->items = new_items;
	list->current = new_current;
	list->last_selected = 0;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, tracklist, rows, n_rows);

	/* FIXME: Free old items */
}


void
bmp_playlist_tracklist_remove_unselected (BmpPlaylist *self, gint tracklist)
{
	GList *new_items;
	BmpListType *list;
	BmpFileItem *file_item;
	gint n, new_count, new_current, n_removed_rows = 0;
	GArray *removed_rows;

	list =
	    bmp_playlist_tracklist_get_i (self, tracklist);
	new_items =
	    NULL;
	new_count =
	    0;
	new_current =
	    -1;
	removed_rows =
	    g_array_new (FALSE, FALSE, sizeof(gint));

	for (n = 0; n < g_list_length(list->items); n++)
	  {
		file_item = LIST_GET_NTH_ITEM (list, n);

		/* fixme: is it really ok to access privately? by doing this be basically say, the function pointers
			  in the list struct are only for the chromalist widget.. */
		if (file_item->selected)
		  {
			new_items = g_list_append (new_items, file_item);
			if (n == list->current)
				new_current = new_count;
			new_count++;
		  }
		else
	          {
			g_array_append_val (removed_rows, n);
			n_removed_rows++;
		  }
	  }

	list->items = new_items;
	list->current = new_current;
	list->last_selected = 0;

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, tracklist, removed_rows, n_removed_rows);

	/* FIXME: Free old items */
}


static gint
tracklist_find_func (gconstpointer a, gconstpointer b)
{
    BmpPlaylistItem *playlist_item = (BmpPlaylistItem*)a;

    if (!a) return 1;

    if (!strcmp(playlist_item->listname, (gchar*)b)) return 0;
    return 1;
}

BmpListType*
bmp_playlist_tracklist_get_i (BmpPlaylist *self, gint _idx)
{
	BmpPlaylistItem *playlist_item;
	GList *llink;

	llink = g_list_nth (self->private->playlists->items, _idx);
	if (! llink)
	    return NULL;

	playlist_item = llink->data;
	if (! playlist_item)
	{
	    g_message("%s: No link data; returning.", G_STRLOC);
	    return NULL;
	}

	return playlist_item->list;
}


BmpListType*
bmp_playlist_tracklist_get (BmpPlaylist *self, const gchar *name)
{
	BmpPlaylistItem *playlist_item;
	GList *llink;

	if (!name)
	    return NULL;

	llink = g_list_find_custom (self->private->playlists->items, name, tracklist_find_func);
	if (!llink)
	    return NULL;

	playlist_item = llink->data;
	if (!playlist_item)
	    return NULL;

	return playlist_item->list;
}

gint
bmp_playlist_tracklist_new (BmpPlaylist *self, const gchar *name)
{
	BmpListType *list;
	gint rv;

	list = bmp_playlist_tracklist_get (self, name);
	if (list)
	  {
	    g_message ("%s: List exists; returning", G_STRFUNC);
	    return -1;
          }

	list = g_new0(BmpListType,1);

	list->name = g_strdup (name);
	list->column_data_func = (ColumnDataFunc)tracklist_column_data_func;
	list->row_get_selected_func = (RowGetSelectedFunc)tracklist_get_selected_func;
	list->row_set_selected_func = (RowSetSelectedFunc)tracklist_set_selected_func;
	list->on_display_func = (OnDisplayFunc)tracklist_on_display_func;
	list->item_swap_func = (ItemSwapFunc)tracklist_item_swap_func;
	list->item_destroy_func = (GDestroyNotify)tracklist_item_destroy_func;
	list->list_sort_func = (ListSortFunc)tracklist_sort_func;
	list->set_dirty_func = (SetDirtyFunc)tracklist_set_dirty_func;
	list->get_dirty_func = (GetDirtyFunc)tracklist_get_dirty_func;
	list->set_rendered_func = (SetRenderedFunc)tracklist_set_rendered_func;
	list->get_rendered_func = (GetRenderedFunc)tracklist_get_rendered_func;
	list->items = NULL;
	list->current = 0;
	list->multiple_selected = FALSE;
	list->last_selected = -1;
	list->adjustment = 0.0;
	list->time = time(NULL);

	rv = insert_playlist_item (self, name, list);

        g_signal_emit (G_OBJECT (self),
                   signals[BMP_PLAYLIST_SIGNAL_LIST_ADDED],
                   0, rv);

	return rv;
}

static void
save_playlist (BmpPlaylistItem *playlist_item, BmpPlaylist *self)
{
	BmpFileItem *item;
	GList *iter_list;
	GError *error = NULL;
	gchar *buffer = NULL, *aux;

	iter_list = playlist_item->list->items;

	while (iter_list)
	{
	    gchar *filename;

	    item =
		iter_list->data;
	    filename =
		g_filename_from_uri (item->uri, NULL, NULL);

	    if ((buffer!=NULL) && filename && g_file_test (filename, G_FILE_TEST_EXISTS))
	      {
			g_free (filename);
			aux = g_strconcat (buffer, "\n", item->uri, NULL);
			g_free (buffer);
			buffer = g_strdup(aux);
			g_free (aux);
	      }
	    else if (filename && g_file_test (filename, G_FILE_TEST_EXISTS))
	      {
			buffer = g_strconcat (item->uri, NULL);
	      }

	    iter_list = g_list_next(iter_list);
	}

	if ( (! buffer) || (! strlen(buffer))) {
		    buffer = g_strdup("");
	}

	g_file_set_contents (playlist_item->filename, buffer, -1, &error);

	g_free (buffer);

	if (error) {
		g_message("SAVE playlist [%s] ERROR: %s", playlist_item->filename, error->message);
	}
}

void
bmp_playlist_save_lists (BmpPlaylist *self)
{
	const gchar *name;

	g_return_if_fail (BMP_IS_PLAYLIST (self));

	if (!bmp_configuration)
	    return;

	if (g_list_length(self->private->playlists->items) == 0) return;
	g_list_foreach (self->private->playlists->items, (GFunc) save_playlist, self);

	/* FIXME: This doesn't belong here, it's an UI issue */
	name = bmp_playlist_get_current_tracklist_name (self);
	bmp_configuration_key_set_string (bmp_configuration,
		"bmp",
		"current-playlist",
		name);
#if 0
	list =
	    bmp_playlist_tracklist_get (self, name);
	if (list)
	    bmp_configuration_key_set_int (bmp_configuration,
		"bmp",
		"current-playlist-position",
		list->current);
#endif
}

void
bmp_playlist_load_playlists (BmpPlaylist *self)
{
	GDir *dir;
	GError *error = NULL;
	gchar *fqname, *listname;
	const gchar* filename;
	gint current_list = 0;
	gint list;

	g_return_if_fail (BMP_IS_PLAYLIST (self));

	dir = g_dir_open (bmp_configuration_key_fetch (bmp_configuration, "bmp", "playlists-path"), 0, &error);

	if (error)
	{
		g_message("LOAD PLAYLISTS ERORR: %s", error->message);
		return;
	}

	while (42)
	{
		filename = g_dir_read_name (dir);
		if (filename == NULL) break;
		if (str_has_suffix_nocase(filename, "m3u"))
		{
			fqname = g_build_filename (bmp_paths[BMP_PATH_PLAYLISTS], filename, NULL);
			load_playlist (self, fqname);
			g_free (fqname);
			current_list++;
		}
	}
	g_dir_close (dir);

	if (! current_list) {
		listname =
		    g_strdup ("Default");
		list =
		    bmp_playlist_tracklist_new (self, listname);

		if (list != -1) {
			bmp_playlist_set_pre_current_name (self, listname);
			bmp_playlist_set_pre_current_position (self, -1);
			bmp_configuration_key_set_string (bmp_configuration, "bmp", "current-playlist", "Default");
		} else
			g_free (listname);
	}

	self->private->playlists->list_sort_func (self, NULL);
}

gint
bmp_playlist_get_current_track_idx (BmpPlaylist *self)
{
	BmpListType *list;
	GList *llink;
	gpointer data;

	if (! self->private->playlists->items)
	    return -1;
	llink = g_list_nth (self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	if (! llink)
	    return -1;

	data  = llink->data;
	if (! data)
	    return -1;
	list  = ((BmpPlaylistItem*)data)->list;
	if (! list)
	    return -1;

	if (! list->items)
	    return -1;

	if (list->current < 0)
	    return -1;

	return list->current;
}

gchar*
bmp_playlist_get_current_uri (BmpPlaylist *self)
{
	BmpListType *list;
	BmpFileItem *item;
	GList *llink;
	gpointer data;
	gchar *uri;

	if (! self->private->playlists->items)
	  {
	    g_message("%s: No Items. Returning NULL.", G_STRLOC);
	    return NULL;
	  }
	llink = g_list_nth (self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	if (! llink)
	  {
	    g_message("%s: No llink. Returning NULL.", G_STRLOC);
	    return NULL;
	  }
	data  = llink->data;
	if (! data)
	  {
	    g_message("%s: No llink data. Returning NULL.", G_STRLOC);
	    return NULL;
	  }
	list  = ((BmpPlaylistItem*)data)->list;
	if (! list)
	  {
	    g_message("%s: No list. Returning NULL.", G_STRLOC);
	    return NULL;
	  }

	if (! list->items)
	  {
	    g_message("%s: No list items. Returning NULL.", G_STRLOC);
	    return NULL;
	  }

	if (list->current < 0)
	  {
	    g_message("%s: List current < 0. Returning NULL.", G_STRLOC);
	    return NULL;
	  }

	item = g_list_nth(list->items, list->current)->data;
	if (!item)
	  {
	    g_message("%s: No item. Returning NULL.", G_STRLOC);
	    return NULL;
	  }

	uri = g_strdup (item->uri);

	return uri;
}

gchar*
bmp_playlist_get_current_title (BmpPlaylist *self)
{
	BmpListType *list;
	BmpFileItem *item;
	GList *llink;
	gpointer data;
	gchar *tracknumber, *trackname, *title;
	const gchar *listname;

	if (! self->private->playlists->items)
	    return NULL;
	llink = g_list_nth (self->private->playlists->items, g_object_get_int (bmp_system_control, "active_tracklist"));
	if (! llink)
	    return NULL;
	data  = llink->data;
	if (! data)
	    return NULL;
	list  = ((BmpPlaylistItem*)data)->list;
	if (! list)
	    return NULL;

	if (! list->items)
	    return NULL;

	if (list->current < 0)
	    return NULL;

	item = g_list_nth(list->items, list->current)->data;
	if (! item)
	    return NULL;

	tracklist_on_display_func (item);

	trackname = bmp_metadata_get_titlestring (item->metadata);
	tracknumber = g_hash_table_lookup (item->metadata, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	listname = bmp_playlist_get_nth_tracklist_name (bmp_playlist, g_object_get_int (bmp_system_control, "active_tracklist"));

	title = g_strdup_printf("[%s] %d. %s", listname, list->current+1, trackname);

	return title;
}
